# -*- coding: utf-8 -*-
import jieba, pymongo, numpy, math, pyprind
from collections import defaultdict
from ngram import NGram
from itertools import dropwhile
from functools import reduce
from udicOpenData.stopwords import rmsw
from udic_nlp_API.settings import W2VMODEL, DEBUG

if DEBUG:
    import requests
else:
    from KCM.__main__ import KCM

class WikiKCEM(object):
    """docstring for WikiKCEM"""
    def __init__(self, uri=None):
        if not DEBUG:
            self.kcmObject = KCM('cht', 'cht', uri=uri)

        self.client = pymongo.MongoClient(uri)['nlp']
        self.Collect = self.client['wiki']
        self.reverseCollect = self.client['wikiReverse']
        self.kcem = self.client['kcem']
        
        self.model = W2VMODEL
        self.wikiNgram = NGram((i['key'] for i in self.reverseCollect.find({}, {'key':1, '_id':False})))
        self.kcemNgram = NGram((i['key'] for i in self.kcem.find({}, {'key':1, '_id':False})))
            
    @staticmethod
    def findPath(keyword):
        Collect = pymongo.MongoClient(None)['nlp']['wikiReverse']
        cursor = Collect.find({'key':keyword}).limit(1)[0]
        if 'ParentOfLeafNode' in cursor:
            cursor = cursor['ParentOfLeafNode']
        else:
            cursor = cursor['parentNode']

        queue = {(parent, (keyword, parent,)) for parent in set(cursor) - set(keyword)}

        while queue:
            (keyword, path) = queue.pop()
            cursor = Collect.find({'key':keyword}).limit(1)
            if cursor.count():
                parentNodes = cursor[0]
                parentNodes = parentNodes['parentNode']
                for parent in parentNodes:
                    if parent in path:
                        yield path[:path.index(parent) + 1]
                    else:
                        queue.add((parent, path + (parent, )))
            else:
                yield path

    def get(self, keyword):
        # if not in wiki Category
        # use wiki Ngram to search
        result = self.kcem.find({'key':keyword}, {'_id':False}).limit(1)
        if result.count():
            return result[0]
        else:
            kcemKeyword = self.kcemNgram.find(keyword)
            if kcemKeyword:
                result = self.kcem.find({'key':kcemKeyword}, {'_id':False}).limit(1)[0]
                return {**result, 'similarity':self.kcem.compare(keyword, kcemKeyword)}
            return {'key':keyword, 'value':[], 'similarity':0}

    def buildParent(self, keyword, denominator=None):
        cursor = self.reverseCollect.find({'key':keyword}, {'_id':False}).limit(1)
        if not cursor.count(): return
        cursor = cursor[0].get('ParentOfLeafNode', []) + cursor[0].get('parentNode', [])
        candidate = {parent: self.toxinomic_score(keyword, parent, denominator) for parent in cursor}

        # 如果parent有多種選擇， 那就把跟keyword一模一樣的parent給剔除
        # 但是如果parent只有唯一的選擇而且跟關鍵字 一樣那就只能選他了
        if len(candidate) > 1 and keyword in candidate:
            del candidate[keyword]

        # Use Min-max normalization
        # 因為最後輸出的值為機率，而機率不能是負的
        # 所以先透過min-max轉成0~1的數值範圍
        M, m = max(candidate.items(), key=lambda x:x[1])[1], min(candidate.items(), key=lambda x:x[1])[1]
        if M == m or len(candidate) == 0:
            M, m = 1, 0

        summation = 0
        for k, v in candidate.items():
            candidate[k] = (v - m) / (M - m)
    
            # 算機率
            summation += candidate[k]

        # calculate possibility, and the sum of possibility is 1.
        if summation:
            for k, v in candidate.items():
                candidate[k] = v / summation
        # 意思：key=lambda x:(-x[1], len(x[0]))
        # 先以分數排序，若同分則優先選擇字數少的category當答案
        result = {'key':keyword, 'value':sorted(candidate.items(), key=lambda x:(-x[1], len(x[0]))), 'similarity':1}
        self.kcem.insert(result)
        return result

    def toxinomic_score(self, keyword, parent, denominator):
        jiebaCut = jieba.lcut(parent, cut_all=True)
        def similarityScore():
            def getSimilarity(keyword, term):
                try:
                    similarity = self.model.similarity(keyword, term)
                    # sign = 1 if similarity > 0 else -1
                    # return sign * (similarity ** 2)
                    return similarity ** 2
                except KeyError as e:
                    return 0
            scoreList = [getSimilarity(keyword, term) for term in jiebaCut]
            return reduce(lambda x, y: x+y, scoreList)

        def kcmScore():
            if DEBUG:
                keywordKcm = dict(requests.get('http://140.120.13.244:10000/kcm/?keyword={}&lang=cht&num=-1'.format(keyword)).json()['value'])
            else:
                keywordKcm = dict(self.kcmObject.get(keyword, -1).get('value', []))
            if keywordKcm:
                keywordKcmTotal = sum(keywordKcm.values())
                return reduce(lambda x,y:x+y, [(keywordKcm.get(term, 0) / keywordKcmTotal)**2 for term in jiebaCut])
            else:
                return 0

        def harmonic_mean():
            cosine, kcm = similarityScore() , kcmScore()
            if cosine and kcm:
                return 2 * (cosine/len(jiebaCut) * kcm/len(jiebaCut)) / (cosine/len(jiebaCut) + kcm/len(jiebaCut))
            else:
                return 0

        if denominator:
            return (similarityScore() + kcmScore()) / denominator(jiebaCut)
        else:
            # default mode
            # return (similarityScore() + kcmScore()) / 1.17 ** len(jiebaCut)
            return harmonic_mean()

    def classify(self, file):
        import json
        kcemTable = {}
        context = ''.join(i['context'] for i in json.load(open(file, 'r')))
        result = defaultdict(list)
        # for key in rmsw(context, 'n'):
        for key in pyprind.prog_bar(list(rmsw(context, 'n'))):
            rawParent = self.get(key)
            parent = kcemTable.setdefault(key, rawParent['value'][0][0] if rawParent['value'] and rawParent['key'] == key else None)
            if parent:
                result[parent].append(key)
        result = {k:len(v) for k,v in result.items()}
        return sorted(result.items(), key=lambda x:-x[1])


    def loss(self):
        import json
        import numpy as np
        answer = json.load(open('answer.json', 'r'))
        power = []
        for i in np.arange(1.0, 1.5, 0.005):
            power.append(sum([dict(self.get(term, lambda jiebaCut:i ** len(jiebaCut))['value'])[answer[term]] for term in answer]))
        print(power)
        print('1 + lnN')
        print(sum([dict(self.get(term, lambda jiebaCut:len(jiebaCut) * math.log1p(len(jiebaCut)))['value'])[answer[term]] for term in answer]))
        print('1 + log10 N')
        print(sum([dict(self.get(term, lambda jiebaCut:1+math.log(len(jiebaCut), 10))['value'])[answer[term]] for term in answer]))
        print('1 + log2 N')
        print(sum([dict(self.get(term, lambda jiebaCut:1+math.log(len(jiebaCut), 2))['value'])[answer[term]] for term in answer]))
        print('N (1 + logN)')
        print(sum([dict(self.get(term, lambda jiebaCut: len(jiebaCut)*(1+math.log(len(jiebaCut), 2)))['value'])[answer[term]] for term in answer]))
        print('1 + NlogN')
        print(sum([dict(self.get(term, lambda jiebaCut:1+len(jiebaCut) * math.log(len(jiebaCut), 2))['value'])[answer[term]] for term in answer]))

        print('1 + NlogX N')
        power = []
        for i in np.arange(1.05, 10, 0.05):
            power.append(sum([dict(self.get(term, lambda jiebaCut:1+len(jiebaCut)*math.log(len(jiebaCut), i))['value'])[answer[term]] for term in answer]))
        print(power)

        print('N (1 + logX N)')
        power = []
        for i in np.arange(1.05, 10, 0.05):
            power.append(sum([dict(self.get(term, lambda jiebaCut: len(jiebaCut)*(1+math.log(len(jiebaCut), i)))['value'])[answer[term]] for term in answer]))
        print(power)


if __name__ == '__main__':
    import argparse
    """The main routine."""
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description='''
        build kcm model and insert it into mongoDB with command line.    
    ''')
    parser.add_argument('-k', metavar='keyword', help='keyword')
    parser.add_argument('-t', help='test which benchmark is better')
    parser.add_argument('-c', help='use classfiy mode')
    parser.add_argument('-f', help='file used to classfiy')
    args = parser.parse_args()
    wiki = WikiKCEM()
    if args.k:
        while True:
            keyword = input('\nplease input the keyword you want to query:\n')
            keyword = keyword.encode('utf-8').decode('utf-8', errors='ignore')
            print(wiki.get(keyword))
    if args.t:
        wiki.loss()
    if args.c:
        print(wiki.classify(args.f))


from django.core.management.base import BaseCommand, CommandError
from udic_nlp_API.settings_database import uri
from kcem import WikiKCEM
import multiprocessing, pymongo, logging, threading, math

logging.basicConfig(format='%(levelname)s : %(asctime)s : %(message)s', filename='buildKCEM.log', level=logging.INFO)
class Command(BaseCommand):
    help = 'use this to build kcem!'
    
    def build(self):
        k = WikiKCEM(uri)
        keywordList = [i['key'] for i in self.Query.find({}, {'key':1, '_id':False})]
        step = math.ceil(len(keywordList)/multiprocessing.cpu_count())
        keywordPieces = [keywordList[i:i + step] for i in range(0, len(keywordList), step)]
        logging.info('start building kcem')
        self.Collect.remove({})

        def activateKCEM(keywordThreadList):
            ThreadResult = []
            for index, keyword in enumerate(keywordThreadList):
                ThreadResult.append(k.findParent(keyword))
                if index % 1000 == 0:
                    logging.info("已處理 %d 個單子" % index)
                    self.Collect.insert(ThreadResult)
                    ThreadResult = []

        workers = [threading.Thread(target=activateKCEM, kwargs={'keywordThreadList':keywordPieces[i]}, name=str(i)) for i in range(multiprocessing.cpu_count())]

        for thread in workers:
           thread.start()

        # Wait for all threads to complete
        for thread in workers:
            thread.join()
        self.Collect.create_index([("key", pymongo.HASHED)])

    def handle(self, *args, **options):
        self.client = pymongo.MongoClient(uri)
        self.db = self.client['nlp']
        self.Query = self.db['wikiReverse']
        self.Collect = self.db['kcem']
        self.build()
        self.stdout.write(self.style.SUCCESS('build kcem model success!!!'))


from django.core.management.base import BaseCommand, CommandError
from udic_nlp_API.settings_database import uri
import pymongo

class Command(BaseCommand):
    help = 'use this to test kcem!'

    def handle(self, *args, **options):
        client = pymongo.MongoClient(uri)['nlp']
        reverseCollect = client['wikiReverse']
        Collect = client['wiki']

        dictionary = {i['key'] for i in reverseCollect.find()}
        dictionary.update({i['key'] for i in Collect.find()})
        final = list(filter(lambda x:'template:' not in x and 'talk:' not in x and 'user:' not in x and 'category talk:' not in x and 'wikipedia:' not in x and 'draft:' not in x and 'portal:' not in x and 'wikipedia:' not in x and 're:start' not in x and '模塊:' not in x and 'help:' not in x, dictionary))

        with open('wiki.dict.txt', 'w') as f:
            for i in final:
                f.write('{} 999 n\n'.format(i))

        self.stdout.write(self.style.SUCCESS('dump wiki dictionary success!!!'))


# author: Shane Yu  date: April 8, 2017
from django.core.management.base import BaseCommand, CommandError
import pyprind, json
from gensim import models
from utils import criteria
from django.http import HttpRequest
from kcem.views import kcem as kcemRequest

class Command(BaseCommand):
    help = 'use this to test kcem!'
    
    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument(
            '--kcemMode',
            default='hybrid',
            type=str,
            help='mode of kcem, you can choose one from hybrid, w2v or kcem.',
        )
        parser.add_argument(
            '--ans',
            default='Ontology_from_google.json',
            type=str,
            help="a json file, which we crawled from Google's Ontology",
        )
        parser.add_argument(
            '--output',
            default='kcem.loss.json',
            type=str,
            help='output file of kcem loss.',
        )
        parser.add_argument(
            '--upperbound',
            default='5',
            type=int,
            help='Upper bound of kcm and kem parameters.',
        )

    def main(self, kcmNum, kemNum):
        loss = 0
        total = 0

        httpReq = HttpRequest()
        httpReq.method = 'GET'
        httpReq.GET['lang'] = 'cht'
        httpReq.GET['kcm'] = kcmNum
        httpReq.GET['kem'] = kemNum

        for key, ans in self.data:
            httpReq.GET['keyword'] = key

            myans = json.loads(kcemRequest(httpReq).getvalue().decode('utf-8'))
            if myans:
                myans = criteria(self.mode, myans, key)[0][0]
                try:
                    loss += ((1-float(self.model.similarity(myans, ans)))*10)**2
                    total += 1
                except Exception as e:
                    print(e)
                    continue
        return loss, total

    def handle(self, *args, **options):
        self.model = models.KeyedVectors.load_word2vec_format('med400.model.bin', binary=True)
        self.data = json.load(open(options['ans'], 'r')).items()
        self.mode = options['kcemMode']
        file = {}

        for kcmNum in pyprind.prog_bar(range(2, options['upperbound'], 2)):
            for kemNum in range(2, options['upperbound'], 2):
                loss, total = self.main(kcmNum, kemNum)
                print("finish {} test, total loss is {}".format(total, loss / total))
                file['{}-{}'.format(kcmNum, kemNum)] = loss / total

        show = sorted(file.items(), key=lambda x:x[1])
        print('min loss is {}, kcm-kem is : {}'.format(show[0][1], show[0][0]))
        json.dump(file, open(options['output'],'w'))

        self.stdout.write(self.style.SUCCESS('build kem model success!!!'))


# Create your tests here.
from django.test import TestCase
from kcem.crawler.crawler import WikiCrawler

class OrderTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        wiki = WikiCrawler()

    def test_weird_wiki_page(self):
        assert wiki.crawl('馬爾維納斯羣島') == None
        assert wiki.crawl('xbox%20360遊戲封面') == None
        assert wiki.crawl('特色級時間條目') == None
        wiki.checkMissing()
        wiki.mergeMongo()
        
    def test_normal_page(self):
        assert wiki.crawl('中式麵條') == None
        wiki.checkMissing()
        wiki.mergeMongo()


from django.core.management.base import BaseCommand, CommandError
from kcem.crawler.crawler import WikiCrawler

class Command(BaseCommand):
    help = 'use this to crawl Wikipedia !'
    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument('--crawl', type=bool)
        parser.add_argument('--fix', type=bool)
        parser.add_argument('--disambiguation', type=bool)
    
    def handle(self, *args, **options):
        wiki = WikiCrawler()
        if options['crawl']:
            wiki.CrawlFromDumpData()
            wiki.crawl('頁面分類')
            wiki.crawl('日本動畫師')
            wiki.crawl('特色級時間條目')
            # wiki.crawl('中式麵條')
            wiki.crawl('各国动画师')
            # wiki.crawl('中央大学校友')
            # wiki.crawl('媒體')
            wiki.crawl('xbox%20360遊戲封面')
            # wiki.crawl('喜欢名侦探柯南的维基人')
            # wiki.crawl('日本原創電視動畫')
            # wiki.crawl('富士電視台動畫')
            # wiki.crawl('萌擬人化')
            # wiki.checkMissing()
            wiki.mergeMongo()
        elif options['fix']:
            wiki.checkMissing()
            wiki.mergeMongo()
        elif options['disambiguation']:
            wiki.dfs('全部消歧義頁面', disambiguation=True)
        self.stdout.write(self.style.SUCCESS('crawl Wikipedia success!!!'))


# -*- coding: utf-8 -*-
import requests, os.path, threading, multiprocessing, pymongo, logging, sys, subprocess, json_lines, os, urllib
from bs4 import BeautifulSoup
from collections import defaultdict
from opencc import OpenCC
from pyquery import PyQuery
from kcem.ignorelist import IGNORELIST
from pyquery import PyQuery as pq
from udic_nlp_API.settings_database import uri

# develop = True
develop = False
if develop:
    database = 'test'
else:
    database = 'nlp'

class WikiCrawler(object):
    """docstring for WikiCrawler"""
    def __init__(self):
        self.logName = 'WikiCrawler.log'
        logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO, filename=self.logName)
        self.openCC = OpenCC('s2t')
        self.wikiBaseUrl = 'https://zh.wikipedia.org'
        self.client = pymongo.MongoClient(uri)[database]
        self.Collect = self.client['wiki']
        self.reverseCollect = self.client['wikiReverse']
        self.queueLock = threading.Lock()
        self.visited, self.stack, self.urlStack = set(), [], []
            
    def crawl(self, root):
        self.stack.append(root)
        # 為了避免stack一開始就太空導致沒有工作做，先把stack塞多點工作
        self.dfs()
        self.dfs()
        self.dfs()
        self.dfs()
        self.dfs()

        self.thread_init(self.thread_job, self.dfs)

    def thread_init(self, thread_job, thread_func):
        logging.info("thread init start")
        workers = [threading.Thread(target=thread_job, args=(thread_func, ), name=str(i)) for i in range(multiprocessing.cpu_count())]
        for thread in workers:
           thread.start()
        # Wait for all threads to complete
        for thread in workers:
            thread.join()
        logging.info('finish init')

    def thread_job(self, thread_func):
        while True:
            # if return value of thread_func is True
            # means stack is empty.
            if thread_func():
                break
        logging.info("finish thread job") 

    @staticmethod
    def genUrl(category, disambiguation=False):
        if disambiguation:
            return 'https://zh.wikipedia.org/zh-tw/' + urllib.parse.quote(category)
        return 'https://zh.wikipedia.org/zh-tw/Category:' + urllib.parse.quote(category)

    def dfs(self, parent=None, disambiguation=False):
        try:
            if not parent:
                self.queueLock.acquire()
                if self.stack:
                    parent = self.stack.pop()
                    self.queueLock.release()
                else:
                    self.queueLock.release()
                    logging.info("stack is empty!!")
                    # means the end of thread job
                    return True

            result = defaultdict(dict)
    
            res = requests.get(self.genUrl(parent)).text
            soup = BeautifulSoup(res, 'lxml')
    
            def push_2_stack(tradText):
                if tradText not in self.visited and True not in {i in tradText for i in IGNORELIST}:
                    self.stack.append(tradText)
    
            def node():
                # node
                nonlocal result
    
                # 代表這是一個重導向的頁面
                # 需要特殊邏輯去爬
                # e.q. https://zh.wikipedia.org/zh-tw/Category:馬爾維納斯羣島
                redirect = soup.select('#SoftRedirect a')
                if redirect:
                    childNode = redirect[0].text.replace('Category:', '')
                    result[parent].setdefault('node', []).append(childNode)
                    push_2_stack(childNode)
                    return

                    
    
                for childNode in soup.select('.CategoryTreeLabelCategory'):
                    tradText = childNode.text
                    
                    # build dictionary
                    result[parent].setdefault('node', []).append(tradText)
    
                    # if it's a node hasn't been through
                    # append to stack
                    push_2_stack(tradText)
    
            def leafNode():
                def disAmb(childText):
                    disAmbResult = {'key':childText, 'leafNode':[]}
                    soup = pq(self.genUrl(childText, disambiguation=True))
                    if soup('#參考來源, #参考来源'):
                        disambiguationCandidates = soup('#參考來源, #参考来源').parent().prev_all().items()
                    elif soup('#參見, #参见'):
                        disambiguationCandidates = soup('#參見, #参见').parent().prev_all().items()
                    else:
                        disambiguationCandidates = soup('#mw-content-text a').items()
                    disambiguationCandidates = {pq(j).text() for i in disambiguationCandidates for j in i('a').filter(lambda x: pq(this).attr('href') and pq(this).text() and '/wiki/' in pq(this).attr('href') and ':' not in pq(this).attr('href') and pq(this).text() not in IGNORELIST)}
                    for candidate in disambiguationCandidates:
                        disAmbResult['leafNode'].append(candidate)
                    self.Collect.insert(disAmbResult)

                nonlocal result

                # leafNode (要注意wiki的leafNode有下一頁的連結，都要traverse完)
                leafNodeList = [soup.select('#mw-pages a')]
                while leafNodeList:
                    current = leafNodeList.pop(0)
    
                    # notyet 變數的意思是，因為wiki會有兩個下一頁的超連結
                    # 頂部跟底部
                    # 所以如果把頂部的bs4結果append到leafNodeLIst的話
                    # 底部就不用重複加
                    notyet = True
                    for child in current:
                        childText = child.text
                        print(childText)
                        if notyet and childText in ('下一頁', '下一页') and child.has_attr('href'):
                            notyet = False
                            print(self.wikiBaseUrl + child['href'])
                            leafNodeList.append(BeautifulSoup(requests.get(self.wikiBaseUrl + child['href']).text, 'lxml').select('#mw-pages a'))
                        elif childText not in ['下一頁', '上一頁', '下一页', '上一页']:
                            result[parent].setdefault('leafNode', []).append(childText)

                            if disambiguation:
                                disAmb(childText)
                # insert
                result = [dict({'key':key}, **value) for key, value in result.items()]

            def grandParent():
                nonlocal result
                # 空的頁面，啥都沒有...
                # 這個時候就直接幫他把自己當作leafNode結束這個分支吧
                # e.q. https://zh.wikipedia.org/zh-tw/Category:特色級時間條目
                if not result:
                    result.append({'key': parent, 'leafNode':[parent]})

                # grandParent Node
                # 每個頁面的最底下都會說他們parent category是什麼
                # 也把它爬其來避免爬蟲有時候因為網路錯誤而漏掉
                parentOfNode = soup.select('#mw-normal-catlinks li a')
                if not parentOfNode:
                    logging.error("no grand parent: {}".format(self.genUrl(parent)))
                else:
                    for gdParent in parentOfNode:
                        result.append({'key':gdParent.text, 'node':[parent]})

            node()
            leafNode()
            grandParent()
            
            if result:
                # [BUG] pymongo.errors.DocumentTooLarge: BSON document too large (39247368 bytes) - the connected server supports BSON document sizes up to 16777216 bytes.
                # 會噴錯的都是大陸省份奇怪的資料，不處理並不影響效能
                self.Collect.insert(result)
            else:
                logging.error("no result: {}".format(self.genUrl(parent)))
            self.visited.add(parent)
            logging.info('now is at {} url:{} result:{}'.format(parent, self.genUrl(parent), result))
        except Exception as e:
            logging.error('{} has occured an Exception. \n {}'.format(parent, e))
            raise e

    def checkMissing(self):
        def grepErrorFromLog():
            logging.info('start grepErrorFromLog')
            with open(self.logName, 'r') as f:
                result = set()
                for i in f:
                    if ': ERROR :' in i:
                        if 'no grand parent: ' in i:
                            url = i.split('no grand parent: ')[1]
                        else:
                            url = i.split('no result: ')[1]
                        url = ''.join(url.rsplit('\n', 1))
                        parent = url.rsplit(':', 1)[-1]
                        result.add(parent)

            self.stack.extend(result)
            # emtpy log file
            open(self.logName, 'w').close()
            logging.info('end grepErrorFromLog')

        self.visited = set()
        while True:
            self.stack = []

            # 把WikiCrawler.log裏面發生error的頁面都再爬過一遍
            # 出錯的原因千百種，許多原因不明，那些頁面都是正常的wiki頁面
            # 少部份是因為該wiki為特殊頁面，如重導向頁面、圖片檔庫存頁等等，導致爬蟲邏輯並不適用
            grepErrorFromLog()

            # 把mongo所有點都看過，如果有一個名詞出現在node陣列中，但是在wiki裏面卻查不到該名詞，代表dfs因為不知名原因沒有繼續鑽進去爬，所以現在挑錯就把他挑出來，繼續爬
            for index, child in enumerate(self.Collect.find({}, {'_id':False})):
                self.visited.add(child['key'])
                if 'node' in child:
                    for node in child["node"]:
                        if not self.Collect.find({'key':node}).limit(1).count():
                            self.stack.append(node)
            logging.warning('missing part is : {}'.format(self.stack))
            logging.info('miss : {}'.format(str(len(self.stack) / index)))
            if not self.stack:
                # 經過再三檢查，確定沒有漏掉的node沒有繼續dfs進去後，就可以結束了
                break
            # 把以前沒爬到的都放進stack中之後就起動thread去爬吧
            self.thread_init(self.thread_job, self.dfs)

    def mergeMongo(self):
        def buildInvertedIndex():
            # build reverseCollect
            reverseResult = defaultdict(dict)
            for parent in self.Collect.find({}, {'_id':False}):
                for leafNode in parent.get('leafNode', []):
                    reverseResult[leafNode].setdefault('ParentOfLeafNode', []).append(parent['key'])
                for node in parent.get('node', []):
                    reverseResult[node].setdefault('parentNode', []).append(parent['key'])
            reverseResult = [dict({'key':key}, **value) for key, value in reverseResult.items()]
            self.reverseCollect.remove({})
            self.reverseCollect.insert(reverseResult)
            self.reverseCollect.create_index([("key", pymongo.HASHED)])
            logging.info("buildInvertedIndex done")

        # merge Collect
        logging.info("merge begin")
        result = defaultdict(dict)
        for term in self.Collect.find({}, {'_id':False}):
            for key, value in term.items():
                if key != 'key':
                    result[self.openCC.convert(term['key']).lower()].setdefault(key, set()).update({self.openCC.convert(i).lower() for i in value})

        insertList = []
        for key, value in result.items():
            for k, s in value.items():
                value[k] = list(s)
            insertList.append(dict({'key':key}, **value))
        self.Collect.remove({})
        self.Collect.insert(insertList)
        self.Collect.create_index([("key", pymongo.HASHED)])

        logging.info("merge done")

        buildInvertedIndex()



    def CrawlFromDumpData(self):
        logging.info('start download wiki data')
        if not os.path.isfile('build/zhwiki-latest-pages-articles.xml.bz2'):
            subprocess.call(['wget', 'https://dumps.wikimedia.org/zhwiki/latest/zhwiki-latest-pages-articles.xml.bz2'])
        subprocess.call(['WikiExtractor.py', 'zhwiki-latest-pages-articles.xml.bz2', '-o', 'wikijson', '--json'])
        self.urlStack = []
        for (dir_path, dir_names, file_names) in os.walk('.'):
            if 'wikijson' in dir_path:
                for file_name in file_names:
                    with open(os.path.join(dir_path, file_name), 'r') as f:
                        for item in json_lines.reader(f):
                            self.urlStack.append(item['url'])

        def dumpDataFunc():
            try:
                self.queueLock.acquire()
                if self.urlStack:
                    url = self.urlStack.pop()
                    self.queueLock.release()
                else:
                    self.queueLock.release()
                    logging.info("urlStack is empty!!")
                    # means the end of thread job
                    return True

                soup = PyQuery(url)
                parent = soup('#firstHeading').text()
                logging.info('now dumpDataFunc is at {}, url: {} '.format(parent, url))
                result = [{'key':gdParent.text(), 'leafNode':[parent]} for gdParent in soup('#mw-normal-catlinks li a').items()]
                
                if result:
                    # [BUG] pymongo.errors.DocumentTooLarge: BSON document too large (39247368 bytes) - the connected server supports BSON document sizes up to 16777216 bytes.
                    # 會噴錯的都是大陸省份奇怪的資料，不處理並不影響效能
                    self.Collect.insert(result)
                else:
                    logging.error("no result at: {}".format(url))
            except Exception as e:
                logging.error('{} has occured an Exception. \n {}'.format(url, e))
        self.thread_init(self.thread_job, dumpDataFunc)


# -*- coding: utf-8 -*-
from django.http import JsonResponse
from djangoApiDec.djangoApiDec import queryString_required
from kcem import WikiKCEM
from udic_nlp_API.settings_database import uri

k = WikiKCEM(uri)
# Create your views here.
@queryString_required(['keyword'])
def kcem(request):
    """Generate list of term data source files
    Returns:
        if contains invalid queryString key, it will raise exception.
    """
    keyword = request.GET['keyword']
    return JsonResponse(k.get(keyword), safe=False)


from distutils.core import setup

setup(
    name = 'kcem',
    packages = ['kcem'],
    package_data={'kcem':['crawler/*', 'management/commands/*']},
    version = '4.1',
    description = 'kcem class file',
    author = 'davidtnfsh',
    author_email = 'davidtnfsh@gmail.com',
    url = 'https://github.com/udicatnchu/kcem',
    download_url = 'https://github.com/udicatnchu/kcem/archive/v4.1.tar.gz',
    keywords = ['kcem'],
    classifiers = [],
    license='GPL3.0',
    install_requires=[
        'simplejson',
        'requests',
        'pymongo',
        'bs4',
        'lxml',
        'jieba',
        'numpy',
        'ngram',
        'pytest',
        'json-lines',
        'pyquery',
    ],
    dependency_links=[
        'git+git://github.com/yichen0831/opencc-python.git@master#egg=opencc-python',
        'git+git://github.com/attardi/wikiextractor.git@2a5e6aebc030c936c7afd0c349e6826c4d02b871',
    ],
    zip_safe=True
)
